import torch
import torch.multiprocessing as mp
import queue
import logging
from typing import Dict, Any
from aquilesimage.utils import setup_colored_logger
import time

logger = setup_colored_logger("Aquiles-Worker", logging.INFO)

def inference_worker_loop(
    gpu_id: int,
    work_queue: mp.Queue,
    result_queue: mp.Queue,
    model_name: str,
    config: Dict[str, Any]
):
    try:
        torch.cuda.set_device(gpu_id)
        device = f"cuda:{gpu_id}"
        
        logger.info(f"[Worker-{gpu_id}] Starting on {device}")
        logger.info(f"[Worker-{gpu_id}] Model: {model_name}")

        pipeline = _load_pipeline_in_worker(model_name, gpu_id, config)
        
        logger.info(f"[Worker-{gpu_id}] Pipeline loaded successfully")

        result_queue.put({
            'type': 'worker_ready',
            'gpu_id': gpu_id,
            'device': device
        })

        while True:
            try:
                batch_data = work_queue.get(timeout=0.1)
                
                if batch_data is None:
                    logger.info(f"[Worker-{gpu_id}] Shutdown signal received")
                    break

                request_id = batch_data.get('request_id')
                prompts = batch_data.get('prompts')
                images = batch_data.get('images', None)
                params = batch_data.get('params', {})
                
                logger.info(
                    f"[Worker-{gpu_id}] Processing batch: "
                    f"request_id={request_id}, "
                    f"prompts={len(prompts)}, "
                    f"has_images={images is not None}"
                )
                
                start_time = time.time()

                try:
                    if images is not None:
                        output = pipeline(
                            prompt=prompts,
                            image=images,
                            **params
                        )
                    else:
                        output = pipeline(
                            prompt=prompts,
                            **params
                        )
                    
                    inference_time = time.time() - start_time
                    
                    logger.info(
                        f"[Worker-{gpu_id}] Batch completed: "
                        f"request_id={request_id}, "
                        f"time={inference_time:.2f}s, "
                        f"images={len(output.images)}"
                    )

                    result_queue.put({
                        'type': 'inference_result',
                        'request_id': request_id,
                        'images': output.images,
                        'success': True,
                        'inference_time': inference_time,
                        'gpu_id': gpu_id
                    })
                    
                except Exception as inference_error:
                    logger.error(
                        f"[Worker-{gpu_id}] Inference error: {inference_error}",
                        exc_info=True
                    )
                    
                    result_queue.put({
                        'type': 'inference_result',
                        'request_id': request_id,
                        'error': str(inference_error),
                        'success': False,
                        'gpu_id': gpu_id
                    })
                
            except queue.Empty:
                continue
                
            except Exception as e:
                logger.error(
                    f"[Worker-{gpu_id}] Unexpected error in main loop: {e}",
                    exc_info=True
                )
                continue
        
        logger.info(f"[Worker-{gpu_id}] Shutting down gracefully")
        
    except Exception as e:
        logger.error(
            f"[Worker-{gpu_id}] FATAL ERROR during initialization: {e}",
            exc_info=True
        )
        result_queue.put({
            'type': 'worker_error',
            'gpu_id': gpu_id,
            'error': str(e)
        })


def _load_pipeline_in_worker(model_name: str, gpu_id: int, config: Dict[str, Any]):
    from aquilesimage.pipelines import ModelPipelineInit
    from aquilesimage.models import ImageModel

    torch.set_float32_matmul_precision("high")
    
    if torch.cuda.is_available():
        torch.backends.cudnn.benchmark = True
        torch.backends.cuda.matmul.allow_tf32 = True
        torch.backends.cudnn.deterministic = False
        torch.backends.cudnn.allow_tf32 = True

    low_vram = config.get('low_vram', False)
    auto_pipeline = config.get('auto_pipeline', False)
    device_map_flux2 = config.get('device_map')
    
    logger.info(f"[Worker-{gpu_id}] Initializing pipeline...")

    if auto_pipeline:
        initializer = ModelPipelineInit(
            model=model_name,
            auto_pipeline=True,
            dist_inf=False
        )
    elif device_map_flux2 == 'cuda' and model_name == ImageModel.FLUX_2_4BNB:
        initializer = ModelPipelineInit(
            model=model_name,
            device_map_flux2='cuda',
            dist_inf=False
        )
    else:
        initializer = ModelPipelineInit(
            model=model_name,
            low_vram=low_vram,
            dist_inf=False
        )

    model_pipeline = initializer.initialize_pipeline()
    model_pipeline.start()
    
    logger.info(f"[Worker-{gpu_id}] Pipeline started successfully")

    return model_pipeline.pipeline


import torch
import torch.multiprocessing as mp
from typing import List, Dict, Any, Optional
import logging
from aquilesimage.utils import setup_colored_logger
import time

logger = setup_colored_logger("Aquiles-WorkerManager", logging.INFO)


class WorkerManager:
    def __init__(
        self,
        model_name: str,
        config: Dict[str, Any],
        num_workers: Optional[int] = None
    ):
        self.model_name = model_name
        self.config = config

        if num_workers is None:
            self.num_workers = torch.cuda.device_count()
        else:
            self.num_workers = min(num_workers, torch.cuda.device_count())
        
        if self.num_workers == 0:
            raise RuntimeError("No CUDA devices available for workers")
        
        logger.info(f"WorkerManager: {self.num_workers} workers will be spawned")

        self.work_queues: List[mp.Queue] = []
        self.result_queues: List[mp.Queue] = []
        self.worker_processes: List[mp.Process] = []
        self.device_ids: List[str] = []
        
        self._started = False
    
    def start(self):
        if self._started:
            logger.warning("WorkerManager already started")
            return
        
        logger.info("Starting workers...")
        
        from aquilesimage.runtime.worker_process import inference_worker_loop
        
        for gpu_id in range(self.num_workers):
            device = f"cuda:{gpu_id}"
            self.device_ids.append(device)

            work_q = mp.Queue(maxsize=20)
            result_q = mp.Queue()
            
            self.work_queues.append(work_q)
            self.result_queues.append(result_q)
            
            # Spawn worker process
            process = mp.Process(
                target=inference_worker_loop,
                args=(gpu_id, work_q, result_q, self.model_name, self.config),
                daemon=False,
                name=f"InferenceWorker-{gpu_id}"
            )
            
            process.start()
            self.worker_processes.append(process)
            
            logger.info(f"Worker process started: PID={process.pid}, GPU={device}")

        self._wait_for_workers_ready()
        
        self._started = True
        logger.info("All workers ready!")
    
    def _wait_for_workers_ready(self, timeout: float = 300.0):
        logger.info("Waiting for workers to initialize...")
        
        ready_count = 0
        start_time = time.time()
        
        while ready_count < self.num_workers:
            if time.time() - start_time > timeout:
                raise TimeoutError(
                    f"Workers did not initialize within {timeout}s. "
                    f"Ready: {ready_count}/{self.num_workers}"
                )

            for idx, result_q in enumerate(self.result_queues):
                try:
                    msg = result_q.get(timeout=0.1)
                    
                    if msg.get('type') == 'worker_ready':
                        gpu_id = msg.get('gpu_id')
                        logger.info(f"Worker {gpu_id} is ready")
                        ready_count += 1
                    
                    elif msg.get('type') == 'worker_error':
                        gpu_id = msg.get('gpu_id')
                        error = msg.get('error')
                        raise RuntimeError(
                            f"Worker {gpu_id} failed to initialize: {error}"
                        )
                
                except Exception:
                    continue
        
        logger.info(f"All {self.num_workers} workers initialized successfully")
    
    def stop(self):
        if not self._started:
            return
        
        logger.info("Stopping workers...")

        for idx, work_q in enumerate(self.work_queues):
            try:
                work_q.put(None, timeout=1.0)
                logger.info(f"Shutdown signal sent to worker {idx}")
            except Exception as e:
                logger.warning(f"Failed to send shutdown to worker {idx}: {e}")

        for idx, process in enumerate(self.worker_processes):
            try:
                process.join(timeout=10.0)
                if process.is_alive():
                    logger.warning(f"Worker {idx} did not stop, terminating...")
                    process.terminate()
                    process.join(timeout=5.0)
                    if process.is_alive():
                        logger.error(f"Worker {idx} still alive, killing...")
                        process.kill()
                else:
                    logger.info(f"Worker {idx} stopped gracefully")
            except Exception as e:
                logger.error(f"Error stopping worker {idx}: {e}")
        
        self._started = False
        logger.info("All workers stopped")
    
    def get_queues(self):
        return self.work_queues, self.result_queues
    
    def get_device_ids(self):
        return self.device_ids